@@ -6,8 +6,10 @@ import dataclasses
 import abc
 import asyncio
 import datetime
+import hashlib
 import inspect
 import json
+import operator
 import re
 import shlex
 from ast import literal_eval
@@ -28,11 +30,13 @@ from typing import (
     Callable,
     Dict,
     Generic,
+    Iterator,
     List,
     Mapping,
     Optional,
     Sequence,
     Tuple,
+    Type,
     TypeVar,
     Union,
     cast,
@@ -41,7 +45,7 @@ from typing_extensions import Final, Protocol, runtime_checkable
 from yarl import URL
 
 from .tokenizer import Pos, Token, tokenize
-from .types import LocalPath, RemotePath
+from .types import AlwaysT, LocalPath, RemotePath, TaskStatus
 
 
 _T = TypeVar("_T")
@@ -49,7 +53,7 @@ _T = TypeVar("_T")
 
 LiteralT = Union[None, bool, int, float, str]
 
-TypeT = Union[LiteralT, "ContainerT", "MappingT", "SequenceT"]
+TypeT = Union[LiteralT, "ContainerT", "MappingT", "SequenceT", LocalPath, AlwaysT]
 
 
 @runtime_checkable
@@ -63,12 +67,18 @@ class MappingT(Protocol):
     def __getitem__(self, key: LiteralT) -> TypeT:
         ...
 
+    def __iter__(self) -> Iterator[LiteralT]:
+        ...
+
 
 @runtime_checkable
 class SequenceT(Protocol):
     def __getitem__(self, idx: LiteralT) -> TypeT:
         ...
 
+    def __iter__(self) -> Iterator[TypeT]:
+        ...
+
 
 class RootABC(abc.ABC):
     @abc.abstractmethod
@@ -127,7 +137,7 @@ async def alen(ctx: CallCtx, arg: TypeT) -> int:
 
 
 async def akeys(ctx: CallCtx, arg: TypeT) -> TypeT:
-    # Async version of len(), async is required for the sake of uniformness.
+    # Async version of list(), async is required for the sake of uniformness.
     if not isinstance(arg, Mapping):
         raise TypeError(f"keys() requires a mapping, got {arg!r}")
     return list(arg)  # type: ignore  # List[...] is implicitly converted to SequenceT
@@ -139,16 +149,117 @@ async def fmt(ctx: CallCtx, spec: str, *args: TypeT) -> str:
     return spec.format(*args)
 
 
+class JSONEncoder(json.JSONEncoder):
+    def default(self, obj: Any) -> Any:
+        if dataclasses.is_dataclass(obj):
+            return dataclasses.asdict(obj)
+        # Let the base class default method raise the TypeError
+        return super().default(obj)
+
+
 async def to_json(ctx: CallCtx, arg: TypeT) -> str:
-    return json.dumps(arg)
+    return json.dumps(arg, cls=JSONEncoder)
 
 
 async def from_json(ctx: CallCtx, arg: str) -> TypeT:
     return cast(TypeT, json.loads(arg))
 
 
+async def hash_files(ctx: CallCtx, *patterns: str) -> str:
+    hasher = hashlib.new("sha256")
+    flow = ctx.root.lookup("flow")
+    # emulate attr lookup
+    workspace: LocalPath = cast(
+        LocalPath,
+        await AttrGetter(ctx.start, ctx.end, name="workspace").eval(
+            ctx.root, flow, start=ctx.start
+        ),
+    )
+    for pattern in patterns:
+        for fname in sorted(workspace.glob(pattern)):
+            with fname.open("rb") as stream:
+                data = stream.read(256 * 1024)
+                if not data:
+                    break
+                hasher.update(data)
+    return hasher.hexdigest()
+
+
+def _check_has_needs(ctx: CallCtx, *, func_name: str) -> None:
+    try:
+        ctx.root.lookup("needs")
+    except LookupError:
+        raise ValueError(f"{func_name}() is only available inside a task definition")
+
+
+def _get_needs_statuses(root: RootABC) -> Dict[str, TaskStatus]:
+    needs = root.lookup("needs")
+    assert isinstance(needs, MappingT)
+    result: Dict[str, TaskStatus] = {}
+    for dependency in needs:
+        dep_ctx = needs[dependency]
+        result[dependency] = dep_ctx.result  # type: ignore
+    return result
+
+
+async def always(ctx: CallCtx, *args: str) -> AlwaysT:
+    _check_has_needs(ctx, func_name="success")
+    return AlwaysT()
+
+
+async def success(ctx: CallCtx, *args: str) -> bool:
+    # When called without arguments, checks that all dependency task
+    # succeeded. If arguments are passed, they should be strings that
+    # name some task from `needs:` field. Function will return true
+    # if all of those tasks succeded.
+    _check_has_needs(ctx, func_name="success")
+    needs_statuses = _get_needs_statuses(ctx.root)
+    if not args:
+        return all(status == TaskStatus.SUCCEEDED for status in needs_statuses.values())
+    else:
+        if not all(isinstance(arg, str) for arg in args):
+            raise ValueError("success() function only accept string arguments")
+        try:
+            return all(needs_statuses[need] == TaskStatus.SUCCEEDED for need in args)
+        except KeyError as e:
+            raise ValueError(
+                "success() function got argument, that"
+                f'is not a defined as dependency: "{e.args[0]}"'
+            )
+
+
+async def failure(ctx: CallCtx, *args: str) -> bool:
+    # When called without arguments, checks if any of dependency task
+    # failed. If arguments are passed, they should be strings that
+    # name some task from `needs:` field. Function will return true
+    # if any of those tasks failure.
+    _check_has_needs(ctx, func_name="failure")
+    needs_statuses = _get_needs_statuses(ctx.root)
+    if not args:
+        return any(status == TaskStatus.FAILED for status in needs_statuses.values())
+    else:
+        if not all(isinstance(arg, str) for arg in args):
+            raise ValueError("failure() function only accept string arguments")
+        try:
+            return any(needs_statuses[need] == TaskStatus.FAILED for need in args)
+        except KeyError as e:
+            raise ValueError(
+                "failure() function got argument, that"
+                f'is not a defined as dependency: "{e.args[0]}"'
+            )
+
+
 FUNCTIONS = _build_signatures(
-    len=alen, nothing=nothing, fmt=fmt, keys=akeys, to_json=to_json, from_json=from_json
+    len=alen,
+    nothing=nothing,
+    fmt=fmt,
+    keys=akeys,
+    to_json=to_json,
+    from_json=from_json,
+    success=success,
+    failure=failure,
+    hash_files=hash_files,
+    always=always,
 )
 
 
@@ -269,7 +380,8 @@ class Call(Item):
     async def eval(self, root: RootABC) -> TypeT:
         args = [await a.eval(root) for a in self.args]
         try:
-            tmp = await self.func.call(root, *args)  # type: ignore
+            call_ctx = CallCtx(self.start, self.end, root)
+            tmp = await self.func.call(call_ctx, *args)  # type: ignore
         except asyncio.CancelledError:
             raise
         except EvalError:
@@ -312,6 +424,62 @@ def make_text(arg: Token) -> Text:
     return Text(arg.start, arg.end, arg.value)
 
 
+@dataclasses.dataclass(frozen=True)
+class BinOp(Item):
+    op: Callable[[TypeT, TypeT], TypeT]
+    left: Item
+    right: Item
+
+    async def eval(self, root: RootABC) -> TypeT:
+        left_val = await self.left.eval(root)
+        right_val = await self.right.eval(root)
+        return self.op(left_val, right_val)  # type: ignore
+
+
+def make_bin_op_expr(args: Tuple[Item, Token, Item]) -> BinOp:
+    op_map = {
+        "==": operator.eq,
+        "!=": operator.ne,
+        "or": operator.or_,
+        "and": operator.and_,
+        "<": operator.lt,
+        "<=": operator.le,
+        ">": operator.gt,
+        ">=": operator.ge,
+    }
+    op_token = args[1]
+    return BinOp(
+        args[0].start,
+        args[2].end,
+        op=op_map[op_token.value],
+        left=args[0],
+        right=args[2],
+    )
+
+
+@dataclasses.dataclass(frozen=True)
+class UnaryOp(Item):
+    op: Callable[[TypeT], TypeT]
+    operand: Item
+
+    async def eval(self, root: RootABC) -> TypeT:
+        operand_val = await self.operand.eval(root)
+        return self.op(operand_val)  # type: ignore
+
+
+def make_unary_op_expr(args: Tuple[Token, Item]) -> UnaryOp:
+    op_map = {
+        "not": operator.not_,
+    }
+    op_token = args[0]
+    return UnaryOp(
+        args[0].start,
+        args[1].end,
+        op=op_map[op_token.value],
+        operand=args[1],
+    )
+
+
 def a(value: str) -> Parser:
     """Eq(a) -> Parser(a, a)
 
@@ -325,6 +493,8 @@ COMMA: Final = skip(a(","))
 
 OPEN_TMPL: Final = skip(a("${{"))
 CLOSE_TMPL: Final = skip(a("}}"))
+OPEN_TMPL2: Final = skip(a("$[["))
+CLOSE_TMPL2: Final = skip(a("]]"))
 
 LPAR: Final = skip(a("("))
 RPAR = skip(a(")"))
@@ -332,6 +502,9 @@ RPAR = skip(a(")"))
 LSQB: Final = skip(a("["))
 RSQB = skip(a("]"))
 
+BIN_OP = a("==") | a("!=") | a("or") | a("and") | a("<") | a("<=") | a(">") | a(">=")
+UNARY_OP = a("not")
+
 REAL: Final = literal("REAL") | literal("EXP")
 
 INT: Final = literal("INT") | literal("HEX") | literal("OCT") | literal("BIN")
@@ -365,13 +538,16 @@ FUNC_ARGS: Final = maybe(EXPR + many(COMMA + EXPR)) >> make_args
 FUNC_CALL: Final = (NAME + LPAR + FUNC_ARGS + RPAR + TRAILER) >> make_call
 
 
-ATOM_EXPR.define(ATOM | FUNC_CALL | LOOKUP)
+ATOM_EXPR.define(ATOM | FUNC_CALL | LOOKUP | LPAR + EXPR + RPAR)
+
+BIN_OP_EXPR: Final = ATOM_EXPR + BIN_OP + EXPR >> make_bin_op_expr
+
+UNARY_OP_EXPR: Final = UNARY_OP + EXPR >> make_unary_op_expr
+
+EXPR.define(BIN_OP_EXPR | UNARY_OP_EXPR | ATOM_EXPR)
 
 
-EXPR.define(ATOM_EXPR)
-
-
-TMPL: Final = OPEN_TMPL + EXPR + CLOSE_TMPL
+TMPL: Final = (OPEN_TMPL + EXPR + CLOSE_TMPL) | (OPEN_TMPL2 + EXPR + CLOSE_TMPL2)
 
 TEXT: Final = some(lambda tok: tok.type == "TEXT") >> make_text
 
@@ -380,27 +556,47 @@ PARSER: Final = oneplus(TMPL | TEXT) + skip(finished)
 
 class Expr(Generic[_T]):
     allow_none = True
+    allow_expr = True
+    type: Type[_T]
+    _ret: Union[None, _T]
+    _pattern: Union[None, str, _T]
+    _parsed: Optional[Sequence[Item]]
 
     @classmethod
-    def convert(cls, arg: str) -> _T:
-        # implementation for StrExpr and OptStrExpr
-        return cast(_T, arg)
+    @abc.abstractmethod
+    def convert(cls, arg: Union[str, _T]) -> _T:
+        pass
 
-    def __init__(self, start: Pos, end: Pos, pattern: Optional[str]) -> None:
+    def _try_convert(self, arg: Union[str, _T], start: Pos, end: Pos) -> None:
+        try:
+            self._ret = self.convert(arg)
+        except (TypeError, ValueError) as exc:
+            raise EvalError(str(exc), start, end)
+
+    def __init__(self, start: Pos, end: Pos, pattern: Union[None, str, _T]) -> None:
         self._pattern = pattern
         # precalculated value for constant string, allows raising errors earlier
-        self._ret: Optional[_T] = None
+        self._ret = None
         if pattern is not None:
-            if not isinstance(pattern, str):
+            if isinstance(pattern, str):
+                # parse later
+                pass
+            elif isinstance(pattern, self.type):
+                # explicit non-string value is passed
+                self._try_convert(pattern, start, end)
+                return
+            else:
                 raise EvalError(f"str is expected, got {type(pattern)}", start, end)
             tokens = list(tokenize(pattern, start=start))
-            self._parsed: Optional[Sequence[Item]] = PARSER.parse(tokens)
-            assert self._parsed is not None
+            if tokens:
+                self._parsed = PARSER.parse(tokens)
+            else:
+                self._parsed = [Text(start, end, "")]
+            assert self._parsed
             if len(self._parsed) == 1 and type(self._parsed[0]) == Text:
-                try:
-                    self._ret = self.convert(cast(Text, self._parsed[0]).arg)
-                except (TypeError, ValueError) as exc:
-                    raise EvalError(str(exc), start, end)
+                self._try_convert(cast(Text, self._parsed[0]).arg, start, end)
+            elif not self.allow_expr:
+                raise EvalError(f"Expressions are not allowed in {pattern}", start, end)
         elif self.allow_none:
             self._parsed = None
         else:
@@ -408,7 +604,9 @@ class Expr(Generic[_T]):
 
     @property
     def pattern(self) -> Optional[str]:
-        return self._pattern
+        if self._pattern is None:
+            return None
+        return str(self._pattern)
 
     async def eval(self, root: RootABC) -> Optional[_T]:
         if self._ret is not None:
@@ -468,12 +666,28 @@ class StrictExpr(Expr[_T]):
 # These comprehensive specializations exist mainly for static type checker
 
 
-class StrExpr(StrictExpr[str]):
-    pass
+class StrExprMixin:
+    @classmethod
+    def convert(cls, arg: str) -> str:
+        return arg
 
 
-class OptStrExpr(Expr[str]):
-    pass
+class StrExpr(StrExprMixin, StrictExpr[str]):
+    type = str
+
+
+class OptStrExpr(StrExprMixin, Expr[str]):
+    type = str
+
+
+class SimpleStrExpr(StrExprMixin, StrictExpr[str]):
+    allow_expr = False
+    type = str
+
+
+class SimpleOptStrExpr(StrExprMixin, Expr[str]):
+    allow_expr = False
+    type = str
 
 
 class IdExprMixin:
@@ -490,80 +704,126 @@ class IdExprMixin:
 
 
 class IdExpr(IdExprMixin, StrictExpr[str]):
-    pass
+    type = str
 
 
 class OptIdExpr(IdExprMixin, Expr[str]):
-    pass
+    type = str
+
+
+class SimpleIdExpr(IdExprMixin, StrictExpr[str]):
+    allow_expr = False
+    type = str
+
+
+class SimpleOptIdExpr(IdExprMixin, Expr[str]):
+    allow_expr = False
+    type = str
 
 
 class URIExprMixin:
     @classmethod
-    def convert(cls, arg: str) -> URL:
+    def convert(cls, arg: Union[str, URL]) -> URL:
         return URL(arg)
 
 
 class URIExpr(URIExprMixin, StrictExpr[URL]):
-    pass
+    type = URL
 
 
 class OptURIExpr(URIExprMixin, Expr[URL]):
-    pass
+    type = URL
 
 
 class BoolExprMixin:
     @classmethod
-    def convert(cls, arg: str) -> bool:
+    def convert(cls, arg: Union[str, bool]) -> bool:
+        if isinstance(arg, bool):
+            return arg
         tmp = parse_literal(arg, "a boolean")
         return bool(tmp)
 
 
 class BoolExpr(BoolExprMixin, StrictExpr[bool]):
-    pass
+    type = bool
 
 
 class OptBoolExpr(BoolExprMixin, Expr[bool]):
-    pass
+    type = bool
+
+
+class SimpleBoolExpr(BoolExprMixin, StrictExpr[bool]):
+    allow_expr = False
+    type = bool
+
+
+class SimpleOptBoolExpr(BoolExprMixin, Expr[bool]):
+    allow_expr = False
+    type = bool
+
+
+class EnableExprMixin:
+    @classmethod
+    def convert(cls, arg: Union[AlwaysT, str, bool]) -> Union[bool, AlwaysT]:
+        if isinstance(arg, AlwaysT) or isinstance(arg, bool):
+            return arg
+        if arg == "always()":
+            return AlwaysT()
+        tmp = parse_literal(arg, "a boolean")
+        return bool(tmp)
+
+
+class EnableExpr(EnableExprMixin, StrictExpr[Union[bool, AlwaysT]]):
+    type = bool
+
+
+class OptEnableExpr(EnableExprMixin, Expr[Union[bool, AlwaysT]]):
+    type = bool
 
 
 class IntExprMixin:
     @classmethod
-    def convert(cls, arg: str) -> int:
+    def convert(cls, arg: Union[str, int]) -> int:
+        if isinstance(arg, int):
+            return arg
         tmp = parse_literal(arg, "an integer")
         return int(tmp)  # type: ignore[arg-type]
 
 
 class IntExpr(IntExprMixin, StrictExpr[int]):
-    pass
+    type = int
 
 
 class OptIntExpr(IntExprMixin, Expr[int]):
-    pass
+    type = int
 
 
 class FloatExprMixin:
     @classmethod
-    def convert(cls, arg: str) -> float:
+    def convert(cls, arg: Union[str, float]) -> float:
+        if isinstance(arg, float):
+            return arg
         tmp = parse_literal(arg, "a float")
         return float(tmp)  # type: ignore[arg-type]
 
 
 class FloatExpr(FloatExprMixin, StrictExpr[float]):
-    pass
+    type = float
 
 
 class OptFloatExpr(FloatExprMixin, Expr[float]):
-    pass
+    type = float
 
 
 class OptLifeSpanExpr(OptFloatExpr):
     RE = re.compile(r"^((?P<d>\d+)d)?((?P<h>\d+)h)?((?P<m>\d+)m)?((?P<s>\d+)s)?$")
 
     @classmethod
-    def convert(cls, arg: str) -> float:
+    def convert(cls, arg: Union[str, float]) -> float:
         try:
-            return float(literal_eval(arg))
+            return super(cls, OptLifeSpanExpr).convert(arg)
         except (ValueError, SyntaxError):
+            assert isinstance(arg, str)
             match = cls.RE.match(arg)
             if match is None:
                 raise ValueError(f"{arg!r} is not a life span")
@@ -578,43 +838,43 @@ class OptLifeSpanExpr(OptFloatExpr):
 
 class LocalPathMixin:
     @classmethod
-    def convert(cls, arg: str) -> LocalPath:
+    def convert(cls, arg: Union[str, LocalPath]) -> LocalPath:
         return LocalPath(arg)
 
 
 class LocalPathExpr(LocalPathMixin, StrictExpr[LocalPath]):
-    pass
+    type = LocalPath
 
 
 class OptLocalPathExpr(LocalPathMixin, Expr[LocalPath]):
-    pass
+    type = LocalPath
 
 
 class RemotePathMixin:
     @classmethod
-    def convert(cls, arg: str) -> RemotePath:
+    def convert(cls, arg: Union[str, RemotePath]) -> RemotePath:
         return RemotePath(arg)
 
 
 class RemotePathExpr(RemotePathMixin, StrictExpr[RemotePath]):
-    pass
+    type = RemotePath
 
 
 class OptRemotePathExpr(RemotePathMixin, Expr[RemotePath]):
-    pass
+    type = RemotePath
 
 
 class OptBashExpr(OptStrExpr):
     @classmethod
     def convert(cls, arg: str) -> str:
-        ret = " ".join(["bash", "-euxo", "pipefail", "-c", shlex.quote(arg)])
+        ret = " ".join(["bash", "-euo", "pipefail", "-c", shlex.quote(arg)])
         return ret
 
 
 class OptPythonExpr(OptStrExpr):
     @classmethod
     def convert(cls, arg: str) -> str:
-        ret = " ".join(["python3", "-c", shlex.quote(arg)])
+        ret = " ".join(["python3", "-uc", shlex.quote(arg)])
         return ret
 
 
