@@ -8,6 +8,7 @@
 import dataclasses
 
 import abc
+import enum
 import yaml
 from typing import (
     Any,
@@ -22,6 +23,7 @@ from typing import (
     Type,
     TypeVar,
     Union,
+    cast,
 )
 from yaml.composer import Composer
 from yaml.constructor import ConstructorError, SafeConstructor
@@ -31,12 +33,13 @@ from yaml.resolver import Resolver as BaseResolver
 from yaml.scanner import Scanner
 
 from . import ast
+from .ast import BatchActionOutputs
 from .expr import (
-    EvalError,
     Expr,
     IdExpr,
     OptBashExpr,
     OptBoolExpr,
+    OptEnableExpr,
     OptIdExpr,
     OptIntExpr,
     OptLifeSpanExpr,
@@ -46,6 +49,11 @@ from .expr import (
     OptStrExpr,
     PortPairExpr,
     RemotePathExpr,
+    SimpleIdExpr,
+    SimpleOptBoolExpr,
+    SimpleOptIdExpr,
+    SimpleOptStrExpr,
+    SimpleStrExpr,
     StrExpr,
     URIExpr,
 )
@@ -91,20 +99,6 @@ class BaseConstructor(SafeConstructor):
         return val
 
 
-def parse_bool(ctor: BaseConstructor, node: yaml.MappingNode) -> bool:
-    return ctor.construct_yaml_bool(node)  # type: ignore[no-untyped-call,no-any-return]
-
-
-BaseConstructor.add_constructor("flow:bool", parse_bool)  # type: ignore
-
-
-def parse_str(ctor: BaseConstructor, node: yaml.MappingNode) -> str:
-    return str(ctor.construct_scalar(node))  # type: ignore[no-untyped-call]
-
-
-BaseConstructor.add_constructor("flow:str", parse_str)  # type: ignore
-
-
 def mark2pos(mark: yaml.Mark) -> Pos:
     return Pos(mark.line, mark.column, LocalPath(mark.name))
 
@@ -200,21 +194,6 @@ class IdMapping(SimpleCompound[_T, Mapping[str, Expr[_T]]]):
         return ret
 
 
-KeyT = Union[
-    None,
-    Type[Expr[Any]],
-    SimpleCompound[Any, Any],
-    Type[ast.Kind],
-]
-VarT = Union[
-    None,
-    Expr[Any],
-    Mapping[str, Any],
-    Sequence[Any],
-    ast.Base,
-    ast.Kind,
-]
-
 _AstType = TypeVar("_AstType", bound=ast.Base)
 _CtorType = TypeVar("_CtorType", bound=BaseConstructor)
 
@@ -222,16 +201,19 @@ _CtorType = TypeVar("_CtorType", bound=BaseConstructor)
 def parse_dict(
     ctor: _CtorType,
     node: yaml.MappingNode,
-    keys: Mapping[str, KeyT],
+    keys: Mapping[str, Any],
     res_type: Type[_AstType],
     *,
     ret_name: Optional[str] = None,
     extra: Optional[Mapping[str, Union[str, LocalPath]]] = None,
     preprocess: Optional[
-        Callable[[_CtorType, Dict[str, VarT]], Dict[str, VarT]]
+        Callable[[_CtorType, yaml.MappingNode, Dict[str, Any]], Dict[str, Any]]
     ] = None,
     find_res_type: Optional[
-        Callable[[_CtorType, Type[_AstType], Dict[str, VarT]], Type[_AstType]]
+        Callable[
+            [_CtorType, yaml.MappingNode, Type[_AstType], Dict[str, Any]],
+            Type[_AstType],
+        ]
     ] = None,
 ) -> _AstType:
     if extra is None:
@@ -243,91 +225,134 @@ def parse_dict(
         raise ConstructorError(
             None,
             None,
-            f"expected a mapping node, but found {node_id}",
+            f"expected a mapping node, but found '{node_id}'",
             node.start_mark,
         )
     node_start = mark2pos(node.start_mark)
     node_end = mark2pos(node.end_mark)
 
-    data = {}
+    data = dict(extra)
     for k, v in node.value:
         key = ctor.construct_object(k)  # type: ignore[no-untyped-call]
         if key not in keys:
             raise ConstructorError(
-                f"while constructing a {ret_name}",
+                f"while constructing a '{ret_name}'",
                 node.start_mark,
-                f"unexpected key {key}",
+                f"unexpected key '{key}'",
                 k.start_mark,
             )
-        item_ctor: KeyT = keys[key]
+        item_ctor: Any = keys[key]
         tmp: Any
-        value: VarT
+        value: Any
         if item_ctor is None:
             # Get constructor from tag
             value = ctor.construct_object(v)  # type: ignore[no-untyped-call]
-        elif item_ctor is ast.Kind:
+        elif isinstance(item_ctor, type) and issubclass(item_ctor, enum.Enum):
             tmp = str(ctor.construct_object(v))  # type: ignore[no-untyped-call]
-            value = ast.Kind(tmp)
+            value = item_ctor(tmp)
         elif isinstance(item_ctor, ast.Base):
             assert isinstance(
                 v, ast.Base
-            ), f"[{type(v)}] {v} should be ast.Base derived"
+            ), f"[{type(v)}] '{v}' should be ast.Base derived"
             value = v
         elif isinstance(item_ctor, SimpleCompound):
             value = item_ctor.construct(ctor, v)
         elif isinstance(item_ctor, type) and issubclass(item_ctor, Expr):
-            tmp = str(ctor.construct_object(v))  # type: ignore[no-untyped-call]
+            tmp = ctor.construct_object(v)  # type: ignore[no-untyped-call]
             value = item_ctor(mark2pos(v.start_mark), mark2pos(v.end_mark), tmp)
         else:
             raise ConstructorError(
-                f"while constructing a {ret_name}",
+                f"while constructing a '{ret_name}'",
                 node.start_mark,
-                f"unexpected value tag {v.tag} for key {key}[{item_ctor}]",
+                f"unexpected value tag '{v.tag}' for key '{key}[{item_ctor}]'",
                 k.start_mark,
             )
         data[key] = value
 
     if preprocess is not None:
-        data = preprocess(ctor, dict(data))
+        data = preprocess(ctor, node, dict(data))
     if find_res_type is not None:
-        res_type = find_res_type(ctor, res_type, dict(data))
+        res_type = find_res_type(ctor, node, res_type, dict(data))
         ret_name = res_type.__name__
 
     optional_fields: Dict[str, Any] = {}
-    found_fields = extra.keys() | data.keys() | {"_start", "_end"}
+    found_fields = data.keys() | {"_start", "_end"}
+    field_names = set()
     for f in dataclasses.fields(res_type):
+        field_names.add(f.name)
         if f.name not in found_fields:
             key = f.name
             item_ctor = keys[key]
-            if item_ctor is None:
-                optional_fields[f.name] = None
-            elif isinstance(item_ctor, SimpleCompound):
-                optional_fields[f.name] = None
-            elif isinstance(item_ctor, ast.Base):
-                optional_fields[f.name] = None
+            if (
+                item_ctor is None
+                or isinstance(item_ctor, SimpleCompound)
+                or isinstance(item_ctor, ast.Base)
+            ):
+                if f.metadata.get("allow_none", False):
+                    optional_fields[f.name] = None
+                else:
+                    raise ConstructorError(
+                        f"while constructing a '{ret_name}', "
+                        f"missing mandatory key '{f.name}'",
+                        node.start_mark,
+                    )
             elif isinstance(item_ctor, type) and issubclass(item_ctor, Expr):
                 if not item_ctor.allow_none:
                     raise ConstructorError(
-                        f"while constructing a {ret_name}, "
-                        f"missing mandatory key {f.name}",
+                        f"while constructing a '{ret_name}', "
+                        f"missing mandatory key '{f.name}'",
                         node.start_mark,
                     )
                 optional_fields[f.name] = item_ctor(node_start, node_end, None)
+            elif isinstance(item_ctor, type) and issubclass(item_ctor, enum.Enum):
+                if f.metadata.get("allow_none", False):
+                    optional_fields[f.name] = None
             else:
                 raise ConstructorError(
-                    f"while constructing a {ret_name}, "
-                    f"unexpected {f.name} constructor type {item_ctor!r}",
+                    f"while constructing a '{ret_name}', "
+                    f"unexpected '{f.name}' constructor type '{item_ctor!r}'",
                     node.start_mark,
                 )
+
+    actual_names = found_fields | optional_fields.keys()
+    missing_names = field_names - actual_names
+    sep = ","
+    if missing_names:
+        raise ConstructorError(
+            f"while constructing a '{ret_name}', "
+            f"missing fields '{sep.join(missing_names)}'",
+            node.start_mark,
+        )
+    unknown_names = actual_names - field_names
+    if unknown_names:
+        raise ConstructorError(
+            f"while constructing a '{ret_name}', "
+            f"unexpected fields '{sep.join(missing_names)}'",
+            node.start_mark,
+        )
     return res_type(  # type: ignore[call-arg]
         _start=node_start,
         _end=node_end,
-        **extra,
         **data,
         **optional_fields,
     )
 
 
+# #### Generics ####
+
+
+def parse_cache(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.Cache:
+    return parse_dict(
+        ctor,
+        node,
+        {
+            "strategy": ast.CacheStrategy,
+            "life_span": OptLifeSpanExpr,
+        },
+        ast.Cache,
+    )
+
+
 # #### Project parser ####
 
 
@@ -341,9 +366,7 @@ class ProjectLoader(Reader, Scanner, Parser, Composer, BaseConstructor, BaseReso
         BaseResolver.__init__(self)
 
 
-PROJECT = {
-    "id": None,
-}
+PROJECT = {"id": None}
 
 
 def parse_project_main(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.Project:
@@ -360,50 +383,53 @@ ProjectLoader.add_path_resolver("project:main", [])  # type: ignore
 ProjectLoader.add_constructor("project:main", parse_project_main)  # type: ignore
 
 
+def parse_project_stream(stream: TextIO) -> ast.Project:
+    ret: ast.Project
+    loader = ProjectLoader(stream)
+    try:
+        ret = loader.get_single_data()  # type: ignore[no-untyped-call]
+        assert isinstance(ret, ast.Project)
+        return ret
+    finally:
+        loader.dispose()  # type: ignore[no-untyped-call]
+
+
+def make_default_project(workspace_stem: str) -> ast.Project:
+    return ast.Project(
+        _start=Pos(0, 0, LocalPath("<default>")),
+        _end=Pos(0, 0, LocalPath("<default>")),
+        id=SimpleIdExpr(
+            Pos(0, 0, LocalPath("<default>")),
+            Pos(0, 0, LocalPath("<default>")),
+            workspace_stem.replace("-", "_"),
+        ),
+    )
+
+
 def parse_project(
     workspace: LocalPath, *, filename: str = "project.yml"
 ) -> ast.Project:
     # Parse project config file
-    ret: ast.Project
     config_file = workspace / filename
     try:
         with config_file.open() as f:
-            loader = ProjectLoader(f)
-            try:
-                ret = loader.get_single_data()  # type: ignore[no-untyped-call]
-                assert isinstance(ret, ast.Project)
-                if not ret.id.isidentifier():
-                    raise EvalError(
-                        f"{ret.id!r} is not identifier", ret._start, ret._end
-                    )
-            finally:
-                loader.dispose()  # type: ignore[no-untyped-call]
+            return parse_project_stream(f)
     except FileNotFoundError:
-        ret = ast.Project(
-            _start=Pos(0, 0, LocalPath("<default>")),
-            _end=Pos(0, 0, LocalPath("<default>")),
-            id=workspace.stem.replace("-", "_"),
-        )
-        if not ret.id.isidentifier():
-            raise ValueError(
-                f"{ret.id!r} is not identifier, "
-                "please rename the project folder or "
-                "specify id in project.yml"
-            )
-    return ret
+        return make_default_project(workspace.stem)
 
 
 # #### Flow parser ####
 
 
 class FlowLoader(Reader, Scanner, Parser, Composer, BaseConstructor, BaseResolver):
-    def __init__(self, stream: TextIO) -> None:
+    def __init__(self, stream: TextIO, *, kind: ast.FlowKind) -> None:
         Reader.__init__(self, stream)
         Scanner.__init__(self)
         Parser.__init__(self)
         Composer.__init__(self)
         BaseConstructor.__init__(self)
         BaseResolver.__init__(self)
+        self._kind = kind
 
 
 def parse_volume(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.Volume:
@@ -444,6 +470,8 @@ def parse_image(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.Image:
             "context": OptLocalPathExpr,
             "dockerfile": OptLocalPathExpr,
             "build_args": SimpleSeq(StrExpr),
+            "env": SimpleMapping(StrExpr),
+            "volumes": SimpleSeq(OptStrExpr),
         },
         ast.Image,
     )
@@ -521,6 +549,7 @@ STRATEGY = {
     "matrix": None,
     "fail_fast": OptBoolExpr,
     "max_parallel": OptIntExpr,
+    "cache": None,
 }
 
 
@@ -528,7 +557,7 @@ def parse_strategy(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.Strateg
     return parse_dict(
         ctor,
         node,
-        {"matrix": None, "fail_fast": OptBoolExpr, "max_parallel": OptIntExpr},
+        STRATEGY,
         ast.Strategy,
     )
 
@@ -538,6 +567,10 @@ FlowLoader.add_path_resolver(  # type: ignore
 )
 FlowLoader.add_constructor("flow:strategy", parse_strategy)  # type: ignore
 
+FlowLoader.add_path_resolver(  # type: ignore
+    "flow:cache", [(dict, "tasks"), (list, None), (dict, "strategy"), (dict, "cache")]
+)
+FlowLoader.add_constructor("flow:cache", parse_cache)  # type: ignore
 
 EXEC_UNIT = {
     "title": OptStrExpr,
@@ -550,7 +583,7 @@ EXEC_UNIT = {
     "python": OptPythonExpr,
     "workdir": OptRemotePathExpr,
     "env": SimpleMapping(StrExpr),
-    "volumes": SimpleSeq(StrExpr),
+    "volumes": SimpleSeq(OptStrExpr),
     "tags": SimpleSeq(StrExpr),
     "life_span": OptLifeSpanExpr,
     "http_port": OptIntExpr,
@@ -558,24 +591,28 @@ EXEC_UNIT = {
 }
 
 
-FlowLoader.add_path_resolver(  # type: ignore
-    "flow:bool", [(dict, "jobs"), (dict, None), (dict, "multi")]
-)
-
-
 JOB = {
     "detach": OptBoolExpr,
     "browse": OptBoolExpr,
     "port_forward": SimpleSeq(PortPairExpr),
-    "multi": None,
+    "multi": SimpleOptBoolExpr,
     **EXEC_UNIT,
 }
 
+JOB_ACTION_CALL = {
+    "action": SimpleStrExpr,
+    "args": SimpleMapping(StrExpr),
+}
 
-def select_shells(ctor: BaseConstructor, dct: Dict[str, Any]) -> Dict[str, Any]:
+
+def select_shells(
+    ctor: BaseConstructor, node: yaml.MappingNode, dct: Dict[str, Any]
+) -> Dict[str, Any]:
     found = {k for k in dct if k in ("cmd", "bash", "python")}
     if len(found) > 1:
-        raise ValueError(f"{','.join(found)} are mutually exclusive")
+        raise ConstructorError(
+            f"{','.join(found)} are mutually exclusive", node.start_mark
+        )
 
     bash = dct.pop("bash", None)
     if bash is not None:
@@ -588,13 +625,51 @@ def select_shells(ctor: BaseConstructor, dct: Dict[str, Any]) -> Dict[str, Any]:
     return dct
 
 
-def parse_job(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.Job:
-    return parse_dict(
-        ctor, node, JOB, ast.Job, preprocess=select_shells  # type: ignore[arg-type]
+def select_job_or_action(
+    ctor: BaseConstructor, node: yaml.MappingNode, dct: Dict[str, Any]
+) -> Dict[str, Any]:
+    if "action" in dct:
+        return {k: v for k, v in dct.items() if k in JOB_ACTION_CALL}
+    else:
+        dct2 = {k: v for k, v in dct.items() if k in JOB}
+        return select_shells(ctor, node, dct2)
+
+
+JOB_OR_ACTION = {**JOB, **JOB_ACTION_CALL}
+
+
+def find_job_type(
+    ctor: BaseConstructor,
+    node: yaml.MappingNode,
+    res_type: Type[ast.Base],
+    arg: Dict[str, Any],
+) -> Union[Type[ast.Job], Type[ast.JobActionCall]]:
+    action = arg.get("action")
+    if action is None:
+        return ast.Job
+    else:
+        return ast.JobActionCall
+
+
+def parse_job(
+    ctor: BaseConstructor, node: yaml.MappingNode
+) -> Union[ast.Job, ast.JobActionCall]:
+    return cast(
+        Union[ast.Job, ast.JobActionCall],
+        parse_dict(
+            ctor,
+            node,
+            JOB_OR_ACTION,
+            ast.Base,
+            preprocess=select_job_or_action,
+            find_res_type=find_job_type,
+        ),
     )
 
 
-def parse_jobs(ctor: BaseConstructor, node: yaml.MappingNode) -> Dict[str, ast.Job]:
+def parse_jobs(
+    ctor: BaseConstructor, node: yaml.MappingNode
+) -> Dict[str, Union[ast.Job, ast.JobActionCall]]:
     ret = {}
     for k, v in node.value:
         key = ctor.construct_id(k)
@@ -611,13 +686,54 @@ TASK = {
     "id": OptIdExpr,
     "needs": SimpleSeq(IdExpr),
     "strategy": None,
+    "enable": OptEnableExpr,
     **EXEC_UNIT,
 }
 
 
-def parse_task(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.Task:
+TASK_ACTION_CALL = {
+    "id": OptIdExpr,
+    "needs": SimpleSeq(IdExpr),
+    "strategy": None,
+    "action": SimpleStrExpr,
+    "args": SimpleMapping(StrExpr),
+}
+
+
+TASK_OR_ACTION = {**TASK, **TASK_ACTION_CALL}
+
+
+def select_task_or_action(
+    ctor: BaseConstructor, node: yaml.MappingNode, dct: Dict[str, Any]
+) -> Dict[str, Any]:
+    if "action" in dct:
+        return {k: v for k, v in dct.items() if k in TASK_ACTION_CALL}
+    else:
+        dct2 = {k: v for k, v in dct.items() if k in TASK}
+        return select_shells(ctor, node, dct2)
+
+
+def find_task_type(
+    ctor: BaseConstructor,
+    node: yaml.MappingNode,
+    res_type: Type[ast.Base],
+    arg: Dict[str, Any],
+) -> Union[Type[ast.Task], Type[ast.TaskActionCall]]:
+    action = arg.get("action")
+    if action is None:
+        return ast.Task
+    else:
+        return ast.TaskActionCall
+
+
+def parse_task(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.Base:
     return parse_dict(
-        ctor, node, TASK, ast.Task, preprocess=select_shells  # type: ignore
+        ctor,
+        node,
+        TASK_OR_ACTION,
+        ast.Base,
+        preprocess=select_task_or_action,
+        find_res_type=find_task_type,
     )
 
 
@@ -631,31 +747,51 @@ FlowLoader.add_path_resolver("flow:tasks", [(dict, "tasks")])  # type: ignore
 FlowLoader.add_constructor("flow:tasks", FlowLoader.construct_sequence)  # type: ignore
 
 
-def parse_flow_defaults(
-    ctor: BaseConstructor, node: yaml.MappingNode
-) -> ast.FlowDefaults:
-    return parse_dict(
-        ctor,
-        node,
-        {
-            "tags": SimpleSeq(StrExpr),
-            "env": SimpleMapping(StrExpr),
-            "workdir": OptRemotePathExpr,
-            "life_span": OptLifeSpanExpr,
-            "preset": OptStrExpr,
-        },
-        ast.FlowDefaults,
-    )
+def parse_flow_defaults(ctor: FlowLoader, node: yaml.MappingNode) -> ast.FlowDefaults:
+    if ctor._kind == ast.FlowKind.LIVE:
+        return parse_dict(
+            ctor,
+            node,
+            {
+                "tags": SimpleSeq(StrExpr),
+                "env": SimpleMapping(StrExpr),
+                "workdir": OptRemotePathExpr,
+                "life_span": OptLifeSpanExpr,
+                "preset": OptStrExpr,
+            },
+            ast.FlowDefaults,
+        )
+    elif ctor._kind == ast.FlowKind.BATCH:
+        return parse_dict(
+            ctor,
+            node,
+            {
+                "tags": SimpleSeq(StrExpr),
+                "env": SimpleMapping(StrExpr),
+                "workdir": OptRemotePathExpr,
+                "life_span": OptLifeSpanExpr,
+                "preset": OptStrExpr,
+                "fail_fast": OptBoolExpr,
+                "max_parallel": OptIntExpr,
+                "cache": None,
+            },
+            ast.BatchFlowDefaults,
+        )
+    else:
+        raise ValueError("Unknown kind {ctor._kind}")
 
 
+FlowLoader.add_path_resolver(  # type: ignore
+    "flow:cache", [(dict, "defaults"), (dict, "cache")]
+)
 FlowLoader.add_path_resolver("flow:defaults", [(dict, "defaults")])  # type: ignore
 FlowLoader.add_constructor("flow:defaults", parse_flow_defaults)  # type: ignore
 
 
 FLOW = {
-    "kind": ast.Kind,
-    "id": None,
-    "title": None,
+    "kind": ast.FlowKind,
+    "id": SimpleOptIdExpr,
+    "title": SimpleOptStrExpr,
     "images": None,
     "volumes": None,
     "defaults": None,
@@ -665,20 +801,11 @@ FLOW = {
 }
 
 
-FlowLoader.add_path_resolver("flow:str", [(dict, "title")])  # type: ignore
-FlowLoader.add_path_resolver("flow:str", [(dict, "id")])  # type: ignore
-
-
-FlowLoader.add_path_resolver(  # type: ignore
-    "flow:str", [(dict, "args"), (dict, "default")]
-)
-FlowLoader.add_path_resolver(  # type: ignore
-    "flow:str", [(dict, "args"), (dict, "descr")]
-)
+ARGS = {"default": SimpleOptStrExpr, "descr": SimpleOptStrExpr}
 
 
 def parse_arg(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.Arg:
-    return parse_dict(ctor, node, {"default": None, "descr": None}, ast.Arg)
+    return parse_dict(ctor, node, ARGS, ast.Arg)
 
 
 def parse_args(ctor: BaseConstructor, node: yaml.MappingNode) -> Dict[str, ast.Arg]:
@@ -687,11 +814,13 @@ def parse_args(ctor: BaseConstructor, node: yaml.MappingNode) -> Dict[str, ast.A
         key = ctor.construct_id(k)
         if isinstance(v, yaml.ScalarNode):
             default = ctor.construct_object(v)  # type: ignore[no-untyped-call]
+            start = mark2pos(v.start_mark)
+            end = mark2pos(v.end_mark)
             ret[key] = ast.Arg(
-                _start=mark2pos(v.start_mark),
-                _end=mark2pos(v.end_mark),
-                default=str(default) if default is not None else None,
-                descr=None,
+                _start=start,
+                _end=end,
+                default=SimpleOptStrExpr(start, end, default),
+                descr=SimpleOptStrExpr(start, end, None),
             )
         else:
             arg = parse_arg(ctor, v)
@@ -703,33 +832,28 @@ FlowLoader.add_path_resolver("flow:args", [(dict, "args")])  # type: ignore
 FlowLoader.add_constructor("flow:args", parse_args)  # type: ignore
 
 
-def select_kind(ctor: BaseConstructor, dct: Dict[str, Any]) -> Dict[str, Any]:
-    if dct["kind"] == ast.Kind.LIVE:
-        tasks = dct.pop("tasks", None)
-        if tasks is not None:
-            raise ValueError("flow of kind={dct['kind']} cannot have tasks")
-        args = dct.pop("args", None)
-        if args is not None:
-            raise ValueError("flow of kind={dct['kind']} cannot have args")
-    elif dct["kind"] == ast.Kind.BATCH:
-        jobs = dct.pop("jobs", None)
-        if jobs is not None:
-            raise ValueError("flow of kind={dct['kind']} cannot have jobs")
-        del dct["jobs"]
-    else:
-        raise ValueError(f"Unknown kind {dct['kind']} of the flow")
-    return dct
-
-
-def find_res_type(
-    ctor: BaseConstructor, res_type: Type[ast.BaseFlow], arg: Dict[str, VarT]
+def find_flow_type(
+    ctor: BaseConstructor,
+    node: yaml.MappingNode,
+    res_type: Type[ast.BaseFlow],
+    arg: Dict[str, Any],
 ) -> Type[ast.BaseFlow]:
-    if arg["kind"] == ast.Kind.LIVE:
+    kind = arg.get("kind")
+    if kind is None:
+        raise ConstructorError(
+            f"missing mandatory key 'kind'",
+            node.start_mark,
+        )
+
+    if kind == ast.FlowKind.LIVE:
         return ast.LiveFlow
-    elif arg["kind"] == ast.Kind.BATCH:
+    elif kind == ast.FlowKind.BATCH:
         return ast.BatchFlow
     else:
-        raise ValueError(f"Unknown kind {arg['kind']} of the flow")
+        raise ConstructorError(
+            f"unknown kind {kind} of the flow",
+            node.start_mark,
+        )
 
 
 def parse_flow_main(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.BaseFlow:
@@ -738,7 +862,7 @@ def parse_flow_main(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.BaseFl
         node,
         FLOW,
         ast.BaseFlow,
-        find_res_type=find_res_type,
+        find_res_type=find_flow_type,
     )
     return ret
 
@@ -747,33 +871,49 @@ FlowLoader.add_path_resolver("flow:main", [])  # type: ignore
 FlowLoader.add_constructor("flow:main", parse_flow_main)  # type: ignore
 
 
+def parse_live_stream(stream: TextIO) -> ast.LiveFlow:
+    loader = FlowLoader(stream, kind=ast.FlowKind.LIVE)
+    try:
+        ret = loader.get_single_data()  # type: ignore[no-untyped-call]
+        assert isinstance(ret, ast.LiveFlow)
+        assert ret.kind == ast.FlowKind.LIVE
+        return ret
+    finally:
+        loader.dispose()  # type: ignore[no-untyped-call]
+
+
 def parse_live(workspace: LocalPath, config_file: LocalPath) -> ast.LiveFlow:
     # Parse live flow config file
     with config_file.open() as f:
-        loader = FlowLoader(f)
-        try:
-            ret = loader.get_single_data()  # type: ignore[no-untyped-call]
-            assert isinstance(ret, ast.LiveFlow)
-            assert ret.kind == ast.Kind.LIVE
-            return ret
-        finally:
-            loader.dispose()  # type: ignore[no-untyped-call]
+        return parse_live_stream(f)
+
+
+def parse_batch_stream(stream: TextIO) -> ast.BatchFlow:
+    loader = FlowLoader(stream, kind=ast.FlowKind.BATCH)
+    try:
+        ret = loader.get_single_data()  # type: ignore[no-untyped-call]
+        assert isinstance(ret, ast.BatchFlow)
+        assert ret.kind == ast.FlowKind.BATCH
+        return ret
+    finally:
+        loader.dispose()  # type: ignore[no-untyped-call]
 
 
 def parse_batch(workspace: LocalPath, config_file: LocalPath) -> ast.BatchFlow:
     # Parse pipeline flow config file
     with config_file.open() as f:
-        loader = FlowLoader(f)
-        try:
-            ret = loader.get_single_data()  # type: ignore[no-untyped-call]
-            assert isinstance(ret, ast.BatchFlow)
-            assert ret.kind == ast.Kind.BATCH
-            return ret
-        finally:
-            loader.dispose()  # type: ignore[no-untyped-call]
+        return parse_batch_stream(f)
 
 
 def find_workspace(path: Optional[Union[LocalPath, str]]) -> ConfigDir:
+    # Find live config file, starting from path.
+    # Return a project root folder and a config folder.
+    #
+    # If path is a not None -- it is used as starting point, LocalPath.cwd() otherwise.
+    # The lookup searches bottom-top from path dir up to the root folder,
+    # looking for .neuro folder.
+    # If the config folder not found -- raise an exception.
+
     if path is not None:
         if not isinstance(path, LocalPath):
             path = LocalPath(path)
@@ -796,19 +936,281 @@ def find_workspace(path: Optional[Union[LocalPath, str]]) -> ConfigDir:
     return ConfigDir(path, path / ".neuro")
 
 
-def find_live_config(workspace: ConfigDir) -> ConfigPath:
-    # Find live config file, starting from path.
-    # Return a project root folder and a path to config file.
-    #
-    # If path is a file -- it is used as is.
-    # If path is a directory -- it is used as starting point, LocalPath.cwd() otherwise.
-    # The lookup searches bottom-top from path dir up to the root folder,
-    # looking for .neuro folder and ./neuro/live.yml
-    # If the config file not found -- raise an exception.
+# #### Action parser ####
 
-    ret = workspace.config_dir / "live.yml"
-    if not ret.exists():
-        raise ValueError(f"{ret} does not exist")
-    if not ret.is_file():
-        raise ValueError(f"{ret} is not a file")
-    return ConfigPath(workspace.workspace, ret)
+
+class ActionLoader(Reader, Scanner, Parser, Composer, BaseConstructor, BaseResolver):
+    def __init__(self, stream: TextIO) -> None:
+        Reader.__init__(self, stream)
+        Scanner.__init__(self)
+        Parser.__init__(self)
+        Composer.__init__(self)
+        BaseConstructor.__init__(self)
+        BaseResolver.__init__(self)
+
+
+INPUT = {
+    "descr": SimpleOptStrExpr,
+    "default": SimpleOptStrExpr,
+}
+
+
+def parse_action_input(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.Input:
+    ret = parse_dict(
+        ctor,
+        node,
+        INPUT,
+        ast.Input,
+    )
+    return ret
+
+
+def parse_action_inputs(
+    ctor: BaseConstructor, node: yaml.MappingNode
+) -> Dict[str, ast.Input]:
+    ret = {}
+    for k, v in node.value:
+        key = ctor.construct_id(k)
+        value = parse_action_input(ctor, v)
+        ret[key] = value
+    return ret
+
+
+ActionLoader.add_path_resolver("action:inputs", [(dict, "inputs")])  # type: ignore
+ActionLoader.add_constructor("action:inputs", parse_action_inputs)  # type: ignore
+
+OUTPUT = {
+    "descr": SimpleOptStrExpr,
+    "value": OptStrExpr,
+}
+
+
+def parse_action_output(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.Output:
+    ret = parse_dict(
+        ctor,
+        node,
+        OUTPUT,
+        ast.Output,
+    )
+    return ret
+
+
+@dataclasses.dataclass(frozen=True)
+class ParsedActionOutputs(ast.Base):
+    # Temporary container. Mapped to real ast in preprocess_action
+    needs: Optional[Sequence[IdExpr]]
+    values: Optional[Mapping[str, ast.Output]]
+
+
+def parse_action_outputs(
+    ctor: BaseConstructor, node: yaml.MappingNode
+) -> ParsedActionOutputs:
+    values = {}
+    needs = None
+    for k, v in node.value:
+        key = ctor.construct_id(k)
+        if key == "needs" and isinstance(v, yaml.SequenceNode):
+            needs = SimpleSeq(IdExpr).construct(ctor, v)
+        else:
+            value = parse_action_output(ctor, v)
+            values[key] = value
+    return ParsedActionOutputs(
+        _start=mark2pos(node.start_mark),
+        _end=mark2pos(node.end_mark),
+        needs=needs,  # type: ignore[arg-type]
+        values=values,
+    )
+
+
+ActionLoader.add_path_resolver("action:outputs", [(dict, "outputs")])  # type: ignore
+ActionLoader.add_constructor("action:outputs", parse_action_outputs)  # type: ignore
+
+
+def parse_job_in_live_action(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.Job:
+    ret = parse_job(ctor, node)
+    if not isinstance(ret, ast.Job):
+        raise ConstructorError(
+            f"nested actions are forbidden",
+            node.start_mark,
+        )
+    return ret
+
+
+ActionLoader.add_path_resolver("action:job", [(dict, "job")])  # type: ignore
+ActionLoader.add_constructor("action:job", parse_job_in_live_action)  # type: ignore
+
+ActionLoader.add_path_resolver(  # type: ignore
+    "action:matrix",
+    [(dict, "tasks"), (list, None), (dict, "strategy"), (dict, "matrix")],
+)
+ActionLoader.add_constructor("action:matrix", parse_matrix)  # type: ignore
+
+ActionLoader.add_path_resolver(  # type: ignore
+    "action:strategy", [(dict, "tasks"), (list, None), (dict, "strategy")]
+)
+ActionLoader.add_constructor("action:strategy", parse_strategy)  # type: ignore
+
+
+ActionLoader.add_path_resolver("action:cache", [(dict, "cache")])  # type: ignore
+ActionLoader.add_path_resolver(  # type: ignore
+    "action:cache", [(dict, "tasks"), (list, None), (dict, "cache")]
+)
+ActionLoader.add_constructor("action:cache", parse_cache)  # type: ignore
+
+
+ActionLoader.add_path_resolver(  # type: ignore[no-untyped-call]
+    "action:task", [(dict, "tasks"), (list, None)]
+)
+ActionLoader.add_constructor("action:task", parse_task)  # type: ignore
+
+
+ActionLoader.add_path_resolver("action:tasks", [(dict, "tasks")])  # type: ignore
+ActionLoader.add_constructor(  # type: ignore
+    "action:tasks", ActionLoader.construct_sequence
+)
+
+
+def parse_exec_unit(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.ExecUnit:
+    return parse_dict(ctor, node, EXEC_UNIT, ast.ExecUnit, preprocess=select_shells)
+
+
+ActionLoader.add_path_resolver("action:exec", [(dict, "pre")])  # type: ignore
+ActionLoader.add_path_resolver("action:exec", [(dict, "post")])  # type: ignore
+ActionLoader.add_path_resolver("action:exec", [(dict, "main")])  # type: ignore
+ActionLoader.add_constructor("action:exec", parse_exec_unit)  # type: ignore
+
+
+BASE_ACTION = {
+    "name": SimpleOptStrExpr,
+    "author": SimpleOptStrExpr,
+    "descr": SimpleOptStrExpr,
+    "inputs": None,
+    "outputs": None,
+    "kind": ast.ActionKind,
+}
+
+LIVE_ACTION: Dict[str, Any] = {"job": None, **BASE_ACTION}
+
+BATCH_ACTION: Dict[str, Any] = {"cache": None, "tasks": None, **BASE_ACTION}
+
+STATEFUL_ACTION: Dict[str, Any] = {
+    "cache": None,
+    "pre": None,
+    "pre_if": OptBoolExpr,
+    "main": None,
+    "post": None,
+    "post_if": OptBoolExpr,
+    **BASE_ACTION,
+}
+
+ACTION = {
+    **LIVE_ACTION,
+    **BATCH_ACTION,
+    **STATEFUL_ACTION,
+}
+
+
+def preprocess_action(
+    ctor: BaseConstructor, node: yaml.MappingNode, dct: Dict[str, Any]
+) -> Dict[str, Any]:
+    kind = dct.get("kind")
+
+    ret: Dict[str, Any]
+    if kind == ast.ActionKind.LIVE:
+        ret = {k: v for k, v in dct.items() if k in LIVE_ACTION}
+    elif kind == ast.ActionKind.BATCH:
+        ret = {k: v for k, v in dct.items() if k in BATCH_ACTION}
+    elif kind == ast.ActionKind.STATEFUL:
+        ret = {k: v for k, v in dct.items() if k in STATEFUL_ACTION}
+    else:
+        raise ConstructorError(
+            f"missing mandatory key 'kind'",
+            node.start_mark,
+        )
+
+    outputs_tmp: Optional[BatchActionOutputs] = dct.get("outputs")
+    if outputs_tmp and kind != ast.ActionKind.BATCH:
+        if outputs_tmp.needs is not None:
+            raise ConnectionError(
+                f"outputs.needs list is not supported " f"for {kind.value} action kind",
+                node.start_mark,
+            )
+        ret["outputs"] = outputs_tmp.values
+    elif outputs_tmp:
+        if outputs_tmp.needs is None:
+            raise ConnectionError(
+                f"outputs.needs list is required " f"for {kind.value} action kind",
+                node.start_mark,
+            )
+        ret["outputs"] = ast.BatchActionOutputs(
+            _start=outputs_tmp._start,
+            _end=outputs_tmp._end,
+            needs=outputs_tmp.needs,
+            values=outputs_tmp.values,
+        )
+    return ret
+
+
+def find_action_type(
+    ctor: BaseConstructor,
+    node: yaml.MappingNode,
+    res_type: Type[ast.BaseAction],
+    arg: Dict[str, Any],
+) -> Type[ast.BaseAction]:
+    kind = arg.get("kind")
+    if kind is None:
+        raise ConstructorError(
+            f"missing mandatory key 'kind'",
+            node.start_mark,
+        )
+    ret: Type[ast.BaseAction]
+    if kind == ast.ActionKind.LIVE:
+        ret = ast.LiveAction
+    elif kind == ast.ActionKind.BATCH:
+        ret = ast.BatchAction
+    elif kind == ast.ActionKind.STATEFUL:
+        ret = ast.StatefulAction
+    else:
+        raise ConnectionError(f"unknown kind {kind} of the action", node.start_mark)
+    if issubclass(ret, (ast.LiveAction, ast.StatefulAction)):
+        for name, val in arg.get("outputs", {}).items():
+            if val.value.pattern is not None:
+                raise ConnectionError(
+                    f"outputs.{name}.value is not supported "
+                    f"for {kind.value} action kind",
+                    node.start_mark,
+                )
+    return ret
+
+
+def parse_action_main(ctor: BaseConstructor, node: yaml.MappingNode) -> ast.BaseAction:
+    ret = parse_dict(
+        ctor,
+        node,
+        ACTION,
+        ast.BaseAction,
+        preprocess=preprocess_action,
+        find_res_type=find_action_type,
+    )
+    return ret
+
+
+ActionLoader.add_path_resolver("action:main", [])  # type: ignore
+ActionLoader.add_constructor("action:main", parse_action_main)  # type: ignore
+
+
+def parse_action_stream(stream: TextIO) -> ast.BaseAction:
+    ret: ast.Project
+    loader = ActionLoader(stream)
+    try:
+        ret = loader.get_single_data()  # type: ignore[no-untyped-call]
+        assert isinstance(ret, ast.BaseAction)
+        return ret
+    finally:
+        loader.dispose()  # type: ignore[no-untyped-call]
+
+
+def parse_action(action_file: LocalPath) -> ast.BaseAction:
+    # Parse project config file
+    with action_file.open() as f:
+        return parse_action_stream(f)
